/*
 * Copyright 2015 Maximilian Weißböck, Wien
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package n4l.model;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;


public class Module {

	private static final Logger log = Logger.getLogger(Module.class.getName());


	static {
		log.setLevel(Level.WARNING);
	}

	public static final String	OUTDOOR				= "NAModule1";
	public static final String	WIND_GAUGE			= "NAModule2";
	public static final String	RAIN_GAUGE			= "NAModule3";
	public static final String	INDOOR				= "NAModule4";
	public static final String	MAIN_STATION		= "NAMain";

	// temp_trend, pressure_trend
	public static final String	TREND_UP			= "2";
	public static final String	TREND_STABLE		= "1";
	public static final String	TREND_DOWN			= "0";

	// Battery Status for Miniserver
	public static final String	BATTERY_FULL		= "4";
	public static final String	BATTERY_HIGH		= "3";
	public static final String	BATTERY_MEDIUM		= "2";
	public static final String	BATTERY_LOW			= "1";
	public static final String	BATTERY_VERY_LOW	= "0";

	// the station where this module belongs to
	private Station				station;

	private String				name;
	private String				id;
	private String				type;

	// derived from time_utc, this is the time (in seconds) when values where last recorded
	private long				lastUpdate;

	// all the Modules Data are stored in this HashMap
	// this makes it very easy to add ore remove parameters
	SortedMap<String, Object>	values;


	/**
	 * Constructor for Module
	 * 
	 * @param station Base Station from this Module
	 * @param name Name of Station
	 * @param id ID of Station
	 * @param type Type of Station
	 */
	public Module(Station station, String name, String id, String type) {
		this.station = station;
		this.name = name;
		this.id = id;
		this.type = type;
		this.lastUpdate = 0L;
		values = new TreeMap<String, Object>();
	}


	/**
	 * @param key - each key gets the Modules name as prefix when a value is set
	 * @param value
	 */
	public void setValue(String key, Object value) {
		values.put(key, value);
	}


	/**
	 * @return calculate up to date status based on last update time
	 */
	public String getDataUpToDate() {
		Instant instant = Instant.now();
		LocalDateTime ldt = LocalDateTime.ofInstant(instant, ZoneId.of("UTC"));
		ZonedDateTime zdt = ldt.atZone(ZoneId.of("UTC"));
		long now = zdt.toEpochSecond();
		long timeDiff = now - getLastUpdate();

		log.info(() -> "Last update: " + timeDiff + " seconds ago");

		if (timeDiff > 60 * 15) { // 15 minutes
			return "data_up_to_date=0";
		}
		return "data_up_to_date=1";
	}


	/**
	 * Get all the Values concatenated as one key value String
	 * 
	 * @param seperator The separator to use between values, eg " " or "\n"
	 * @return all the Modules values as one long Key-Value String, eg: "key1=1 key2=0 key3=3.3"
	 */
	public String getAllValuesAsKVString(String seperator, boolean withQualifiedName) {
		String prefix = null;
		if (withQualifiedName) {
			prefix = getQualifiedName();
		}
		else {
			prefix = getName();
		}
		prefix += ".";
		
		StringBuilder sb = new StringBuilder();
		for (String key : values.keySet()) {
			Object obj = values.get(key);
			if (obj instanceof String) {
				obj = getValueForName(obj);
			}
			else if (key.toString().equals("battery_vp")) {
				obj = getBatteryStatus(obj);
			}
			else if (key.toString().equals("rf_status")) {
				obj = getRadioStatus(obj);
			}
			else if (key.toString().equals("wifi_status")) {
				obj = getWifiStatus(obj);
			}
			sb.append(prefix).append(key).append("=").append(obj).append(seperator);
		}
		sb.append(prefix).append(getDataUpToDate());
		return sb.toString();
	}

	static final int			RSSI_THRESHOLD_GOOD		= 56;	/* good signal */
	static final int			RSSI_THRESHOLD_MEDIUM	= 71;	/* middle quality signal */
	static final int			RSSI_THRESHOLD_BAD		= 86;	/* bad signal */

	public static final String	WIFI_SIGNAL_GOOD		= "3";
	public static final String	WIFI_SIGNAL_MEDIUM		= "2";
	public static final String	WIFI_SIGNAL_BAD			= "1";
	public static final String	WIFI_SIGNAL_VERY_BAD	= "0";


	private String getWifiStatus(Object obj) {
		int rt = ((Long) obj).intValue();
		if (rt <= RSSI_THRESHOLD_GOOD) return WIFI_SIGNAL_GOOD;
		if (rt <= RSSI_THRESHOLD_MEDIUM) return WIFI_SIGNAL_MEDIUM;
		if (rt <= RSSI_THRESHOLD_BAD) return WIFI_SIGNAL_BAD;
		return WIFI_SIGNAL_VERY_BAD;
	}

	static final int			RADIO_THRESHOLD_FULL	= 60;	/* full signal */
	static final int			RADIO_THRESHOLD_HIGH	= 70;	/* signal high */
	static final int			RADIO_THRESHOLD_MEDIUM	= 80;	/* signal medium */
	static final int			RADIO_THRESHOLD_LOW		= 90;	/* low signal */

	public static final String	RADIO_SIGNAL_FULL		= "4";
	public static final String	RADIO_SIGNAL_HIGH		= "3";
	public static final String	RADIO_SIGNAL_MEDIUM		= "2";
	public static final String	RADIO_SIGNAL_LOW		= "1";
	public static final String	RADIO_SIGNAL_VERY_LOW	= "0";


	private String getRadioStatus(Object obj) {
		int rt = ((Long) obj).intValue();
		if (rt <= RADIO_THRESHOLD_FULL) return RADIO_SIGNAL_FULL;
		if (rt <= RADIO_THRESHOLD_HIGH) return RADIO_SIGNAL_HIGH;
		if (rt <= RADIO_THRESHOLD_MEDIUM) return RADIO_SIGNAL_MEDIUM;
		if (rt <= RADIO_THRESHOLD_LOW) return RADIO_SIGNAL_LOW;
		return RADIO_SIGNAL_VERY_LOW;
	}


	/**
	 * Calculate Battery Status
	 * 
	 * @param obj
	 * @return
	 */
	private String getBatteryStatus(Object obj) {
		int batteryValue = ((Long) obj).intValue();
		switch (type) {
			case OUTDOOR: // fall trough
			case RAIN_GAUGE:
				return outdoorAndRainGaugeBatteryStatus(batteryValue);
			case WIND_GAUGE:
				return windGaugeBatteryStatus(batteryValue);
			case INDOOR:
				return indoorGaugeBatteryStatus(batteryValue);
		}
		log.warning(() -> "Unknown Battery Status for: " + type);
		return "0";
	}

	/* for indoor module */
	/* Battery range: 6000 ... 4200 */
	static final int	INDOOR_BATTERY_LEVEL_FULL	= 5640;	/* full */
	static final int	INDOOR_BATTERY_LEVEL_HIGH	= 5280;	/* high */
	static final int	INDOOR_BATTERY_LEVEL_MEDIUM	= 4920;	/* medium */
	static final int	INDOOR_BATTERY_LEVEL_LOW	= 4560;	/* low */
	/* Below 4560: very low */


	private String indoorGaugeBatteryStatus(int batteryValue) {
		if (batteryValue >= INDOOR_BATTERY_LEVEL_FULL) return BATTERY_FULL;
		if (batteryValue >= INDOOR_BATTERY_LEVEL_HIGH) return BATTERY_HIGH;
		if (batteryValue >= INDOOR_BATTERY_LEVEL_MEDIUM) return BATTERY_MEDIUM;
		if (batteryValue >= INDOOR_BATTERY_LEVEL_LOW) return BATTERY_LOW;
		return BATTERY_VERY_LOW;
	}

	/* for rain gauge and outdoor module */
	/* Battery range: 6000 ... 3600 */
	static final int	BATTERY_LEVEL_FULL		= 5500;	/* full */
	static final int	BATTERY_LEVEL_HIGH		= 5000;	/* high */
	static final int	BATTERY_LEVEL_MEDIUM	= 4500;	/* medium */
	static final int	BATTERY_LEVEL_LOW		= 4000;	/* low */
	/* below 4000: very low */


	private String outdoorAndRainGaugeBatteryStatus(int batteryValue) {
		if (batteryValue >= BATTERY_LEVEL_FULL) return BATTERY_FULL;
		if (batteryValue >= BATTERY_LEVEL_HIGH) return BATTERY_HIGH;
		if (batteryValue >= BATTERY_LEVEL_MEDIUM) return BATTERY_MEDIUM;
		if (batteryValue >= BATTERY_LEVEL_LOW) return BATTERY_LOW;
		return BATTERY_VERY_LOW;
	}

	/* for wind gauge */
	/* Battery range: 6000 ... 3950 */
	static final int	WG_BATTERY_LEVEL_FULL	= 5590;	/* full */
	static final int	WG_BATTERY_LEVEL_HIGH	= 5180;	/* high */
	static final int	WG_BATTERY_LEVEL_MEDIUM	= 4770;	/* medium */
	static final int	WG_BATTERY_LEVEL_LOW	= 4360;	/* low */
	/* below 4360: very low */


	private String windGaugeBatteryStatus(int batteryValue) {
		if (batteryValue >= WG_BATTERY_LEVEL_FULL) return BATTERY_FULL;
		if (batteryValue >= WG_BATTERY_LEVEL_HIGH) return BATTERY_HIGH;
		if (batteryValue >= WG_BATTERY_LEVEL_MEDIUM) return BATTERY_MEDIUM;
		if (batteryValue >= WG_BATTERY_LEVEL_LOW) return BATTERY_LOW;
		return BATTERY_VERY_LOW;
	}


	/**
	 * Miniserver can not handle String data, so we translate string values to numbers
	 * 
	 * @param obj containing String to be translated
	 * @return number for this String
	 */
	private String getValueForName(Object obj) {
		String value = obj.toString();
		switch (value) {
			case "up":
				return TREND_UP;
			case "down":
				return TREND_DOWN;
			case "stable":
				return TREND_STABLE;
		}
		log.warning(() -> "No mapping found for " + getQualifiedName() + "." + value + ", will return -999 instead - this shall be interpreted as an illegal value");
				
		return "-999";
	}


	public String getName() {
		return name;
	}


	public void setName(String name) {
		this.name = name;
	}


	public String getId() {
		return id;
	}


	public void setId(String id) {
		this.id = id;
	}


	public String getType() {
		return type;
	}


	public void setType(String type) {
		this.type = type;
	}


	public void setStation(Station station) {
		this.station = station;
	}


	public Station getStation() {
		return station;
	}


	public long getLastUpdate() {
		return lastUpdate;
	}


	public void setLastUpdate(long lastUpdate) {
		this.lastUpdate = lastUpdate;
	}


	public String getQualifiedName() {
		return getStation().getName() + "." + getName();
	}

	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("Module [station=").append(station.getName()).append(", name=").append(name).append(", id=")
				.append(id).append(", type=").append(type).append("]\n");
		sb.append(getAllValuesAsKVString("\n", false));
		return sb.toString();
	}

}