/*
 * Copyright 2015 Maximilian Weißböck
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package n4l.netatmo;

import java.util.List;
import static java.util.logging.Level.*;
import java.util.logging.Logger;

import org.apache.oltu.oauth2.client.OAuthClient;
import org.apache.oltu.oauth2.client.URLConnectionClient;
import org.apache.oltu.oauth2.client.request.OAuthBearerClientRequest;
import org.apache.oltu.oauth2.client.request.OAuthClientRequest;
import org.apache.oltu.oauth2.client.response.OAuthJSONAccessTokenResponse;
import org.apache.oltu.oauth2.client.response.OAuthResourceResponse;
import org.apache.oltu.oauth2.common.OAuth;
import org.apache.oltu.oauth2.common.exception.OAuthProblemException;
import org.apache.oltu.oauth2.common.exception.OAuthSystemException;
import org.apache.oltu.oauth2.common.message.types.GrantType;
import org.json.JSONObject;

import n4l.Netatmo4Loxone;
import n4l.model.Station;


public class NetatmoHttpClient {

	// API URLs that will be used for requests, see:
	// https://dev.netatmo.com/doc.

	protected final static String			URL_BASE			= "https://api.netatmo.net";
	protected final static String			URL_TOKEN			= URL_BASE + "/oauth2/token";
	protected final static String			URL_DEVICES_LIST	= URL_BASE + "/api/getstationsdata";

	private final OAuthClient				oAuthClient			= new OAuthClient(new URLConnectionClient());

	private String							clientId;
	private String							clientSecret;
	private OAuthJSONAccessTokenResponse	token;
	private long							tokenExpiresAt		= 0;

	private final Logger					log					= Logger.getLogger(NetatmoHttpClient.class.getName());


	/**
	 * Constructor for Client
	 * 
	 * @param clientId
	 * @param clientSecret
	 */
	public NetatmoHttpClient(String clientId, String clientSecret) {
		this.clientId = clientId;
		this.clientSecret = clientSecret;
		log.setLevel(WARNING);
		if (Netatmo4Loxone.getDebugNetatmoHTTPCalss()) {
			log.setLevel(INFO);
		}
	}


	/**
	 * Do the login and save the access token for reuse. If access token still valid, do nothing.
	 * 
	 * @param email
	 * @param password
	 */
	public void login(String email, String password) throws OAuthSystemException, OAuthProblemException {

		if (accessTokenStillValid()) {
			return;
		}

		OAuthClientRequest request = OAuthClientRequest.tokenLocation(URL_TOKEN).setGrantType(GrantType.PASSWORD)
				.setClientId(clientId).setClientSecret(clientSecret).setUsername(email).setPassword(password)
				.setScope("read_station").buildBodyMessage();

		token = oAuthClient.accessToken(request);
		setAccessToken(token);
	}


	/**
	 * Returns all the Station Data for all the Stations owned by this user.
	 * 
	 * See <a href="https://dev.netatmo.com/doc/methods/getstationsdata</a> for more information.
	 * 
	 * @param token The token obtained by the login function.
	 */
	public List<Station> getStationData() throws OAuthSystemException, OAuthProblemException {

		OAuthClientRequest bearerClientRequest = new OAuthBearerClientRequest(URL_DEVICES_LIST)
				.setAccessToken(token.getAccessToken()).buildQueryMessage();
		OAuthResourceResponse resourceResponse = oAuthClient.resource(bearerClientRequest, OAuth.HttpMethod.GET,
				OAuthResourceResponse.class);

		return NetatmoDataExtractor.parseDevicesListAndExtractData(new JSONObject(resourceResponse.getBody()));
	}


	private void setAccessToken(OAuthJSONAccessTokenResponse accessToken) {
		this.token = accessToken;
		// calculate expiration time minus 5 minutes as safety margin
		tokenExpiresAt = System.currentTimeMillis() + accessToken.getExpiresIn() * 1000 - 5 * 60 * 1000;
		log.info(() -> "Get new Access Token, will expire in " + (tokenExpiresAt - System.currentTimeMillis()) / 1000 + " seconds");
	}


	private boolean accessTokenStillValid() {
		// check if accessToken still valid
		if (tokenExpiresAt > System.currentTimeMillis()) {
			log.info("() -> Reuse old Token, will expire in " + (tokenExpiresAt - System.currentTimeMillis()) / 1000 + " seconds");
			return true;
		}
		log.info(() -> "Access Token has expired");
		return false;
	}
}
